from . import components
from .blueprint import simple_component

import os

PRELUDE = [
    "vendor/jquery-3.3.1.min.js",
    "vendor/underscore.js",
    "loader.js",
    "vendor/proxy.min.js",
]

def make_prelude():
    static_folder = simple_component.static_folder
    out = []

    for p in PRELUDE:
        fname = os.path.join(static_folder, p)
        with open(fname) as f:
            out.append(f.read())


    return "\n".join(out)



from .components import *

from .proxy import Proxy, ComponentProxy, HTMLProxy
from .basic import MustacheComponent, BigJSPackage, JSComponent

import flask

class ComponentBridge(CoreComponent, MustacheComponent, BigJSPackage):
    WRAP_COMPONENT = False


class ClientBridge(JSComponent):
    def call(self, fn, *args, **kwargs):
        self.__marshal__()

        t = """
            $C("ComponentBridge", function(m) {
                m.exports.call_on_component("{{id}}", "{{ fn }}", {{ &args }}, {{ &kwargs }});
            });
        """.strip()

        r = pystache.render(t, {
            "fn" : fn,
            "args" : json.dumps(args, default=dump_values),
            "kwargs" : json.dumps(kwargs, default=dump_values),
            "id" : self.__html_id__()
        })

        self.__add_activation__(r)

# A server bridge allows a backbone component to invoke bridge methods on the
# class that inherits from it
class ServerBridge(ClientBridge):
    __remote_calls__ = {}

    @classmethod
    @memoize
    def get_rpc_definitions(cls):
        all = [];


        t = """
    ex.__bridge.{{ fn }} = m.exports.add_invocation("{{ cls }}", "{{ fn }}");
            """.strip()

        for c in cls.__remote_calls__:
            all.append(pystache.render(t, {
                "fn" : c,
                "cls" : cls.__name__
            }))

        return """var ex = exports.default || module.exports; ex.__bridge = {};\n$C("ComponentBridge", function(m) { \n%s\n}\n);""" % ("\n".join(all))

    @classmethod
    @memoize
    def get_js_supplements(cls):
        return [cls.get_rpc_definitions()]

    @classmethod
    def api(cls, fn):
        cls.__remote_calls__[fn.__name__] = fn

    @classmethod
    def replace_refs(cls, obj, refs=None):
        if refs is None:
            refs = []

        if type(obj) == dict:
            if "_H" in obj:
                h = HTMLProxy(obj["_H"])
                refs.append(h)
                return h

            elif "_R" in obj and "_C" in obj:
                c = ComponentProxy(obj["_R"], obj["_C"])
                refs.append(c)
                return c
            elif "_B" in obj and "_C" in obj:
                c = ComponentProxy(obj["_B"], obj["_C"])
                refs.append(c)
                return c

            else:
                for k in obj:
                    obj[k] = cls.replace_refs(obj[k], refs)

        if type(obj) == list:
            return [cls.replace_refs(r, refs) for r in obj]


        return obj


    @classmethod
    def invoke(cls, cid, fn, args=[], kwargs={}):
        # we instantiate a proxy for our class instance here,
        # like:
        refs = []
        args = cls.replace_refs(args, refs)
        kwargs = cls.replace_refs(kwargs, refs)

        for r in refs:
            r.marshal()

        c = ComponentProxy(cid, cls)

        args = [c] + args

        return cls.__remote_calls__[fn](*args, **kwargs), c


mark_virtual(
    ClientBridge,
    ServerBridge,
    ComponentBridge,
    Proxy,
    HTMLProxy
)



from .components import *
from . import proxy

from .. import util

import flask
import dotmap

import assets




class JinjaComponent(Component):
    @classmethod
    @memoize
    def get_template(cls):
        with open(cls.get_file_for_ext("html")) as f:
            return f.read()

    def __render__(self):
        template_str = self.get_template()
        return flask.render_template_string(template_str, **(self.context.toDict()))

class JSComponent(Component):
    ASSET_LOADER=assets.JSAsset
    EXCLUDE_JS = set()

    @classmethod
    @memoize
    def get_js(cls):
        p = cls.get_file_for_ext(cls.ASSET_LOADER.EXT)
        loader = cls.get_asset_loader(p)
        with open(p) as f:
            return loader.transform(f.read())

    @classmethod
    def get_asset_loader(cls, filename):
        loaders = util.inheritors(assets.AssetLoader)
        for l in loaders:
            if l.match(filename):
                return l

        return assets.JSAsset


    @classmethod
    def render_requires(cls, requested):
        cls_dir = cls.BASE_DIR
        from .components import REQUIRE_RE
        def render_requires_for_js(js, basedir):
            requires = REQUIRE_RE.findall(js)
            ret = {}
            for p in requires:
                if p in cls.EXCLUDE_JS:
                    continue

                p = p.strip("'\"")
                loader = cls.get_asset_loader(p)
                jsp = loader.find_file(p, basedir, cls.BASE_DIR)
                js = loader.render_file(jsp)

                if not js:
                    ret[p] = 'console.log("MISSING REQUIRE FILE %s FROM %s");' % (p, cls.__name__)
                    continue

                ret[p] = js
                ret.update(render_requires_for_js(js, os.path.dirname(jsp)))

            return ret

        def render_requires(component, basedir):
            ret = {}

            requires = set(component.get_requires()).intersection(set(requested))

            for p in requires:
                loader = cls.get_asset_loader(p)
                jsp = loader.find_file(p, basedir, cls.BASE_DIR)
                if jsp:
                    js = loader.render_file(jsp)
                    if js:
                        ret[p] = js
                        ret.update(render_requires_for_js(js, os.path.dirname(jsp)))
                    else:
                        ret[p] = 'console.log("MISSING REQUIRE FILE %s FROM %s");' % (p, cls.__name__)
                        continue



            return ret

        return render_requires(cls, cls.__name__)

    def __init__(self, *args, **kwargs):
        self.__marshalled__ = False
        self.client = dotmap.DotMap()
        super(JSComponent, self).__init__(*args, **kwargs)



        self.__activate_str__ = ""
        self.__activations__ = []

    def __add_activation__(self, jscode):
        self.__activations__.append(jscode)

    def __activate_tag__(self):
        self.__activate__()

        a = self.__get_activate_script__()
        if a:
            return jinja2.Markup('<script type="text/javascript">\n%s\n</script>' % a)

        return ""

    def __get_activate_script__(self):
        all = [self.__activate_str__]
        all.extend(self.__activations__)

        return "\n".join(all)

    # override this function to provide a custom activation
    def __activate__(self):
        t = """$C("ComponentBridge", function(m) {
            m.exports.activate_component("{{__html_id__}}", "{{ __template_name__ }}", {{ &__context__ }}, {{ __display_immediately__ }} )
        })"""
        rendered =  pystache.render(t, self)
        self.__activate_str__ = rendered

    def __marshal__(self):
        if not self.__marshalled__:
            flask.request.pudgy.components.add(self)
            self.__marshalled__ = True

    def marshal(self, **kwargs):
        self.client.update(**kwargs)
        self.__marshal__()
        return self

class CSSComponent(Component):
    @classmethod
    @memoize
    def get_css(cls):
        with open(cls.get_file_for_ext("css")) as f:
            data = f.read()
            # wait for CSS
            return cls.add_display_rules(data)

    @classmethod
    def add_display_rules(cls, data):
        return "%s\n .wf_%s, .scoped_%s { display: inherit !important; } " % (data, cls.__name__, cls.__name__)


    def __init__(self, *args, **kwargs):
        super(CSSComponent, self).__init__(self, *args, **kwargs)

        if flask.request:
            if not self.__template_name__ in flask.request.pudgy.css:
                flask.request.pudgy.css.add(self.__template_name__)

class SassComponent(CSSComponent):
    @classmethod
    @memoize
    def get_css(cls):
        css_class = "scope_%s" % (cls.__name__)
        with open(cls.get_file_for_ext("sass")) as f:
            data = f.read()
            data = sass.compile(string=".scoped_%s {\n %s\n }" % (cls.__name__, data))
            return cls.add_display_rules(data)


class MustacheComponent(Component):
    @classmethod
    @memoize
    def get_template(cls):
        with open(cls.get_file_for_ext("mustache")) as f:
            return f.read()

    def __render__(self):
        template_str = self.get_template()
        rendered =  pystache.render(template_str, self.context)
        return rendered

mark_virtual(
    MustacheComponent,
    JSComponent,
    SassComponent,
    CSSComponent,
    JinjaComponent,
)

# A Big Package will automatically include its requires into a
# package definition
class BigJSPackage(JSComponent):
    @classmethod
    def get_defines(cls):
        reqs = cls.get_requires()
        return cls.render_requires(reqs)

class BigCSSPackage(Component):
    pass

mark_virtual(
    BigJSPackage,
    BigCSSPackage,
)



from .components import Component

import flask

from .bigpipe import Pipeline

# for a Page to be a proper Component, it needs to give an ID to its body
class Page(Pipeline):
    def __init__(self, *args, **kwargs):
        super(Page, self).__init__(*args, **kwargs)
        self.__marshal__()

    # TODO: not sure I like having to call super() in activations.
    # they should be different functions, perhaps
    def __activate__(self):
        super(Page, self).__activate__()

        t = '$("body").attr("id", "%s");' % (self.__html_id__())
        self.__add_activation__(t)

class FlaskPage(Page, Pipeline):
    def render(self):
        self.__prepare__()

        kwargs = self.context.toDict()
        r = flask.render_template(self.context.template, **kwargs)
        # notice that we don't call __wrap_div__ on r
        return r



from . import bridge

import pystache
import os

from ..util import memoize, shelve_it

from .components import CoreComponent
from .assets import JSAsset
from .basic import BigJSPackage, JSComponent

import shlex
import subprocess

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

class JSXCompileError(Exception):
    pass

BABEL_PRESETS = set(["@babel/preset-react"])
def get_babel_compiler(presets):
    presets = list(presets)
    if not "@babel/preset-react" in presets:
        presets.append("@babel/preset-react")

    def babel_compile(data, fname='???'):
        preset_str = ",".join(presets)
        cmd = "%s --presets %s -f '%s'" % (BABEL_BIN, preset_str, fname)

        p = subprocess.Popen(shlex.split(cmd),
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout,stderr = p.communicate(data.encode())

        if p.returncode != 0:
            raise JSXCompileError(stdout, stderr)

        return stdout.decode()

    return babel_compile

def dukpy_compile(data, fname='???'):
    # try dukpy as a resort, too, but its slower
    import dukpy
    return dukpy.jsx_compile(data)

BABEL_BIN = os.path.expanduser("./node_modules/.bin/babel")
JSX_COMPILE = get_babel_compiler(presets=["@babel/preset-react"])

if not os.path.exists(BABEL_BIN):
    print("*** COULDNT FIND BABEL BIN (%s), REACT COMPONENTS WONT COMPILE" % (BABEL_BIN))
    print("*** Try setting the babel bin with reactcomponent.set_bin('path/to/babel')")

    JSX_COMPILE = dukpy_compile


class JSXAsset(JSAsset):
    EXT="jsx"

    @classmethod
    @shelve_it("jsx.cache")
    def transform(cls, js):
        return JSX_COMPILE(js, cls.__name__)

# clientside ReactLoader for instantiating react components
# rendered on the server
class ReactLoader(CoreComponent, JSComponent):
    WRAP_COMPONENT = False

class ReactComponent(bridge.ClientBridge):
    EXCLUDE_JS = set(["react", "react-dom", "vendor/react", "vendor/react-dom"])
    ASSET_LOADER=JSXAsset

    @classmethod
    def set_babel_bin(cls, babel_bin):
        global BABEL_BIN
        BABEL_BIN=babel_bin
        if not os.path.exists(babel_bin):
            raise Exception("INVALID BABEL BIN PATH", babel_bin)

    @classmethod
    def set_jsx_compiler(cls, fn):
        global JSX_COMPILE
        JSX_COMPILE = fn

    @classmethod
    def add_babel_presets(cls, *presets):
        BABEL_PRESETS.add(*presets)
        compiler = get_babel_compiler(list(BABEL_PRESETS))
        cls.set_jsx_compiler(compiler)

    @classmethod
    def add_babel_preset(cls, *presets):
        cls.add_babel_presets(*presets)

    def __json__(self):
        self.__marshal__()
        return { "_R" : self.__html_id__() }

    def set_ref(self, name):
        # TODO: validate there is only one of each named ref on the page
        self.__ref__ = name
        return self

    def __activate__(self):
        self.client.update(self.context)

        super(ReactComponent, self).__activate__()

        # we override the activation string with our react activation string
        t = """
            $C("ReactLoader", function(m) {
                m.exports.activate_react_component("{{__html_id__}}", "{{ __template_name__ }}", {{ &__context__ }}, {{ __display_immediately__ }}, "{{ __ref__ }}" )
            });
        """.strip()

        self.__activate_str__ = pystache.render(t, self)

from . import components
components.mark_virtual(ReactComponent, ReactLoader)



from .components import *
from . import proxy

from .. import util

import flask
import dotmap
import json

class AssetLoader(Component):
    EXT="???"
    @classmethod
    def find_file(cls, fname, basedir, cls_dir):
        fname = fname.strip("'\"")
        if fname.endswith(cls.EXT):
            fname,ext = os.path.splitext(fname)

        if fname[0] == ".":
            jsp = "%s.%s" % (os.path.join(cls_dir, basedir, fname), cls.EXT)
        else:
            jsp = "%s.%s" % (os.path.join(cls_dir, fname), cls.EXT)

        return jsp

    @classmethod
    def render_file(cls, filename, error_if_missing=False):
        try:
            with open(filename, "r") as f:
                js = cls.transform(f.read())
                return js
        except IOError as e:
            if error_if_missing:
                raise e

        return ""

    @classmethod
    def match(cls, fname):
        return fname.endswith(".%s" % cls.EXT)

    @classmethod
    def transform(cls, js):
        return js

class JSAsset(AssetLoader):
    EXT="js"


class CssAsset(AssetLoader):
    EXT="css"

    @classmethod
    def inject_css(cls,css):
        return 'var _inj = %s;\n $C._inject_css("", _inj.css);' % json.dumps({ "css": css })

    @classmethod
    def transform(cls, css):
        return cls.inject_css(css)

class SassAsset(CssAsset):
    EXT="sass"

    @classmethod
    def transform(cls, css):
        css = sass.compile(string=css)
        return cls.inject_css(css)



import flask

from .components import Component

STR_CLASS = [str]
try:
    STR_CLASS.append(unicode)
except:
    pass
class Proxy(object):
    def __init__(self, *args, **kwargs):
        pass

class HTMLProxy(Proxy):
    def __init__(self, id, *args, **kwargs):
        self.id = id
        self.__html__ = []
        super(HTMLProxy, self).__init__(*args, **kwargs)

    # jquery is limited to only this component's descendants
    def run_jquery(self, fn, strval, selector=None):
        self.__html__.append((fn, strval, selector));
        return self

    def replace_html(self, val, selector=""):
        self.run_jquery("html", val, selector)

    def append_html(self, val, selector=""):
        self.run_jquery("append", val, selector)

    def marshal(self):
        flask.request.pudgy.components.add(self)

    def get_object(self):
        r = {}
        r["html"] = self.get_html_directives()
        return r

    def get_html_directives(self):
        ret = self.__html__
        self.__html__ = []
        return ret

class ComponentProxy(HTMLProxy):
    def __init__(self, id, cls, *args, **kwargs):
        super(ComponentProxy, self).__init__(id, cls, *args, **kwargs)

        self.id = id

        if type(cls) in STR_CLASS:
            self.component = cls
        elif isinstance(cls, Component) or issubclass(cls, Component):
            self.component = cls.__name__
        else:
            raise Exception("UNKNOWN COMPONENT TO PROXY FOR", cls)

        self.__calls__ = []
        self.__transfer__ = []

    def marshal(self):
        flask.request.pudgy.components.add(self)

    def call(self, fn, *args, **kwargs):
        self.__calls__.append((fn, args, kwargs))
        return self

    def transfer(self, *args):
        self.__transfer__.extend(args)
        return self

    def get_activations(self):
        ret = []
        for t in self.__transfer__:
            ret.append(t.__get_activate_script__())

        self.__transfer__ = []
        return ret

    def get_calls(self):
        r = [ [self.component, self.id] + list(c) for c in self.__calls__ ]
        self.__calls__ = []
        return r

    def get_object(self):
        r = {}
        r["calls"] = self.get_calls()
        r["html"] = self.get_html_directives()
        r["activations"] = self.get_activations()
        return r

    def get_html_directives(self):
        ret = self.__html__
        self.__html__ = []
        return ret





import flask
import jinja2

# JinjaComponent: foo/foo.html
# BackboneComponent: foo/foo.js
# MustacheComponent: foo/foo.mustache
# SassComponent: foo/foo.css (runs it through sass pre-processor)
# CSSComponent:  foo/foo.css (doesn't run it through pre-processor)

# ClientBridge: requires foo/foo.js, exposes the client API on the server
# ServerBridge: requires foo/foo.js, exposes the server API on the client

# FlaskPage: takes template to render as named parameter
# BigJSPackage: renders the component as one large package instead of a split

from .components import Component, set_base_dir, validate_components

# templating
from .basic import MustacheComponent, JinjaComponent
# style
from .basic import SassComponent, CSSComponent, BigCSSPackage
# javascripts
from .basic import JSComponent, BigJSPackage


from .bridge import ClientBridge, ServerBridge
from .bigpipe import Pipeline, Pagelet
from .react import ReactComponent
from .backbone import BackboneComponent
from .page import Page, FlaskPage



from .components import *
from .basic import JSComponent

import flask
import time

class Pipeline(Component):
    def pipeline(self):
        import preparable
        from ..blueprint import marshal_components

        flask.request.pudgy.pipelined = True

        def inject_pagelet(pagelet):
            html = pagelet.render()
            t = "\n<div id='pagelet_%s'><!-- %s --></div>\n" % (pagelet.__html_id__(), flask.escape(html))
            i = "\n<script>$C._inject_pagelet('%s')</script>\n" % pagelet.__html_id__()
            return "\n".join([t, i])


        # This is the main loop for the preparer. It does a sleep spin
        def sleep_spin(preparer):
            start = time.time()
            while True:
                time.sleep(0.01)
                preparer.finish_tasks()

                for p in preparer.finished:
                    if not p.__done__:
                        print("FINISHED", p)
                        p.__done__ = True
                        yield inject_pagelet()

                if preparer.done or not preparer.preparing:
                    break

        def run():
            kwargs = self.context.toDict()
            r = self.render()

            yield r
            yield marshal_components()

            preparer = preparable.Preparer()
            for p in flask.request.pudgy.pagelets:
                if not p.__async__:
                    continue

                p.__done__ = False
                preparer.add(p.__prepare__, [])


            preparer.startup()
            sleep_spin(preparer)

            for p in flask.request.pudgy.pagelets:
                if not p.__async__:
                    continue

                p.__finished__ = True
                flask.request.pudgy.components.add(p)

                yield inject_pagelet(p)

            yield marshal_components()


        return flask.Response(flask.stream_with_context(run()))

class Pagelet(JSComponent):
    def __init__(self, *args, **kwargs):
        self.__finished__ = False
        self.__async__ = False
        super(Pagelet, self).__init__(*args, **kwargs)

        if flask.request:
            flask.request.pudgy.pagelets.add(self)

    def async(self):
        self.__async__ = True

    def render(self):
        if self.__async__ and not self.__finished__:
            t = """<div class='async_pagelet pagelet' id='pl_{{ id }}'> </div>"""
            r = pystache.render(t, { "id" : self.__html_id__() })
            return r
        else:
            return super(Pagelet, self).render()

    def __activate_tag__(self):
        if self.__async__ and not self.__finished__:
            return ""

        return super(Pagelet, self).__activate_tag__()

    def __work__(self):
        pass

mark_virtual(Pagelet, Pipeline)



from __future__ import division, print_function


import pystache
import dotmap
import jinja2
import flask
import json
import sass
import os
import re
import hashlib


REQUIRE_RE = re.compile("""require\(['"](.*?)['"]\)""")
VIRTUAL_COMPONENTS = set()

from ..util import memoize, gethash, inheritors

def dump_values(w):
    if w:
        return w.__json__()

    return None

class Component(object):
    WRAP_COMPONENT = True
    BASE_DIR = None

    @classmethod
    def set_base_dir(cls, base_dir):
        cls.BASE_DIR = base_dir


    @classmethod
    @memoize
    def get_defines(cls):
        pass

    @classmethod
    @memoize
    def get_requires(cls):
        js = cls.get_js()
        requires = REQUIRE_RE.findall(js)

        cleaned =  [r for r in requires if r not in cls.EXCLUDE_JS]
        return cleaned

    @classmethod
    @memoize
    def get_file_for_ext(cls, ext):
        return os.path.join(cls.BASE_DIR, cls.__name__, "%s.%s" % (cls.__name__, ext))

    @classmethod
    @memoize
    def get_css(cls):
        return ""

    @classmethod
    @memoize
    def get_js(cls):
        return ""

    @classmethod
    @memoize
    def get_js_supplements(cls):
        return []

    @classmethod
    @memoize
    def get_template(cls):
        return ""

    @classmethod
    @memoize
    def get_version(cls):
        m = hashlib.md5()
        m.update(cls.get_package().data)

        return m.hexdigest()

    @classmethod
    @memoize
    def get_class(cls):
        return cls.__name__


    @classmethod
    @memoize
    def test_package(cls):
        if cls.__name__ in VIRTUAL_COMPONENTS:
            return

        try:
            pkg = cls.get_package()
        except Exception as e:
            print("ERROR IN PACKAGE", cls.__name__)
            raise e

        return


    @classmethod
    @memoize
    def get_package(cls):
        ret = {}
        t = cls.get_template()
        c = cls.get_css()
        j = cls.get_js()
        js_supplements = cls.get_js_supplements()
        if js_supplements:
            js_supplements = "\n".join(js_supplements)

        r = cls.get_requires()
        d = cls.get_defines()

        ret["template"] = t
        ret["css"] = c
        ret["js"] = "%s\n%s" % (j, js_supplements or "")
        ret["requires"] = r
        ret["defines"] = d

        # clean up items
        ret = {k: v for k, v in ret.items() if v}

        return flask.jsonify(ret)

    def __init__(self, *args, **kwargs):
        self.context = dotmap.DotMap(kwargs)
        self.__template_name__ = str(self.__class__.__name__)
        self.__async__ = False
        self.__hash__ = gethash(self)[:10]



    def __prepare__(self):
        pass

    def __repr__(self):
        return "%s: %x" % (self.__template_name__, id(self))

    def __html_id__(self):
        return "cmp_%s" % self.__hash__

    def __json__(self):
        self.__marshal__()
        return { "_H" : self.__html_id__() }

    def __context__(self):
        return json.dumps(self.client.toDict(), default=dump_values)

    # should we wait for any CSS before revealing the component
    def __display_immediately__(self):
        css = self.get_css()
        if not css:
            return 1

        return 0

    def __marshal__(self):
        return self

    def __html__(self):
        self.__marshal__()
        return self.render()


    def __wrap_div__(self, div):
        if not self.WRAP_COMPONENT:
            return div

        if self.__display_immediately__():
            return "<div id='%s' class='immediate scoped_%s'>%s</div>" % (self.__html_id__(), self.__template_name__, div)

        return "<div id='%s' class='scoped_%s' style='display: none;'>%s</div>" % (self.__html_id__(),
            self.__template_name__, div)

    def __render__(self):
        return ""

    def render(self):
        if not self.__async__: # async components get prepared separately
            self.__prepare__()

        div = self.__render__()
        wrapped = self.__wrap_div__(div)
        return wrapped

class CoreComponent(Component):
    pass

def set_base_dir(d):
    Component.set_base_dir(os.path.join(d, "components"))
    CoreComponent.set_base_dir(os.path.join(d, "core"))

@memoize
def validate_components():
    valid = 0
    broken = []
    virtual_components = []
    for c in inheritors(Component):
        if c.__name__ in VIRTUAL_COMPONENTS:
            virtual_components.append(c)

        try:
            pkg = c.test_package()
            if not c.__name__ in VIRTUAL_COMPONENTS:
                valid += 1
        except Exception as e:
            s = "%s Errors:" % (c.__name__)
            s_ = "-" *  len(s)
            broken.append(c.__name__)
            print(s)
            print(s_)
            print(e)

    print("Validated %s components before first request, %s broken" % (valid + len(broken), len(broken)))
    if broken:
        print("Broken:", ",".join(broken))


def mark_virtual(*cls):
    for c in cls:
        VIRTUAL_COMPONENTS.add(c.__name__)


mark_virtual(
    Component,
    CoreComponent,
)



import hashlib
import time
import os

def inheritors(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses

# from https://stackoverflow.com/questions/16463582/memoize-to-disk-python-persistent-memoization
# https://stackoverflow.com/posts/47385932/revisions
from diskcache import Cache
cache_dir = "./cache/"
def shelve_it(table_name):
    cache_file = os.path.join(cache_dir, table_name)
    d = Cache(cache_file)

    def decorator(func):
        def new_func(*args, **kwargs):
            cache_key = "\r".join([str(arg) for arg in args])

            if cache_key in d:
                return d[cache_key]


            ret = func(*args, **kwargs)
            d[cache_key] = ret

            return ret

        return new_func

    return decorator

def memoize(func):
    cache = {}
    def new_func(*args, **kwargs):
        cache_key = "\r".join([str(arg) for arg in args])

        if cache_key in cache:
            return cache[cache_key]

        ret = func(*args, **kwargs)
        cache[cache_key] = ret
        return ret

    return new_func

def gethash(v):
    m = hashlib.md5()
    t = str("%s" % time.time()).encode("utf-8")
    h = str(hash(v)).encode("utf-8")

    m.update(t)
    m.update(h)

    return m.hexdigest()



from __future__ import print_function

import jinja2
import flask
import json
import dotmap

from flask import Blueprint, render_template, abort

simple_component = Blueprint('components', __name__,
        template_folder='templates', url_prefix='/components',static_folder='static')

from . import prelude, util, components

components.set_base_dir(simple_component.root_path)

import os
import sys

from .util import memoize
from .components import Component

@simple_component.route('/prelude.js')
def get_prelude():
    return prelude.make_prelude()

def get_component_by_name(component):
    all_components = util.inheritors(Component)
    found = None
    for c in all_components:
        if c.__name__ == component:
            found = c
            break

    return found

@simple_component.route("/")
def index():
    pass

@simple_component.route('/<component>/invoke/<fn>', methods=['POST'])
def invoke(component, fn):
    found = get_component_by_name(component)

    js = flask.request.get_json()
    cid = js.get("cid", None)
    args = js.get("args", [])
    kwargs = js.get("kwargs", {})

    err = None
    ret = None

    res = {}
    r = {}
    try:
        ret, proxy = found.invoke(cid, fn, args, kwargs)

        r.update(proxy.get_object())
        r["response"] = ret
    except Exception as e:
        err = str(e)
        r["error"] = err
        print("ERROR INVOKING", component, fn, e)
        import traceback
        traceback.print_exc(file=sys.stdout)

    res[cid] = r

    for p in flask.request.pudgy.components:
        if isinstance(p, components.bridge.Proxy):
            res[p.id] = p.get_object()

    return flask.jsonify(res)


@simple_component.route('/csspkg')
def get_big_css():
    requested = flask.request.args.getlist('components')
    all = []

    for component in requested:
        found = get_component_by_name(component)
        all.append(found.get_css())


    r = flask.Response("\n".join(all))
    r.headers["Content-Type"] = "text/css"
    return r

    abort(404)

@simple_component.route('/<component>/css')
def get_css(component):
    found = get_component_by_name(component)
    if found:
        r = flask.Response(found.get_css())
        r.headers["Content-Type"] = "text/css"
        return r

    abort(404)

@simple_component.route('/<component>/requires')
def get_requires(component):
    requested = flask.request.args.getlist('requires[]')

    found = get_component_by_name(component)
    if found:
        return flask.jsonify(found.render_requires(requested))

    abort(404)

@simple_component.route('/<component>')
def show(component):
    found = get_component_by_name(component)

    if found:
        return found.get_package()

    abort(404)

def add_components():
    flask.request.pudgy = dotmap.DotMap()
    flask.request.pudgy.components = set()
    flask.request.pudgy.css = set()
    flask.request.pudgy.pagelets = set()

def marshal_components(prelude=True):
    from . import components
    # when lc.__html__ is called, __marshal__ is invoked, so we use lc.render()
    # instead
    lc = components.bridge.ComponentBridge()
    lc.context.components = flask.request.pudgy.components
    flask.request.pudgy.components = set()
    html = lc.render()

    # TODO: dont send same component version down multiple times
    component_versions = {}
    for c in lc.context.components:
        if c.__marshalled__:
            component_versions[c.get_class()] = "%s" % c.get_version()
            if c.get_class() in flask.request.pudgy.css:
                flask.request.pudgy.css.remove(c.get_class())

    component_versions[lc.get_class()] = lc.get_version()

    postfix = ""
    if flask.request.pudgy.components:
        postfix = marshal_components(prelude=False)

    css = flask.request.pudgy.css
    flask.request.pudgy.css = set()

    big_package = []
    big_package.sort(key=lambda w: w.__name__)
    per_package = []
    for c in css:
        if type(c) == str:
            found = get_component_by_name(c)
            if isinstance(found, components.BigCSSPackage) or \
                    issubclass(found, components.BigCSSPackage):
                big_package.append(c)
            else:
                per_package.append(c)



    return jinja2.Markup(render_template("inject_components.html",
        css_package=big_package, css_singles=per_package,
        postfix=postfix, prelude=prelude, html=html, url_for=dated_url_for,
        versions=json.dumps(component_versions)))

def render_component(name, **kwargs):
    found = get_component_by_name(name)

    return found(**kwargs)

APP=None
def handle_request_to(endpoint, **values):
    url = flask.url_for(endpoint, **values)
    req = flask.Request.from_values(path=url)
    ctx = APP.test_request_context(path=url)

    with ctx:
        res = APP.dispatch_request()
    return res

def dated_url_for(endpoint, **values):
    # we know prelude always returns a string, so
    # we use python string hashing
    if endpoint == 'components.get_prelude':
        res = handle_request_to(endpoint, **values)
        hsh = hash(res)

        values['q'] = "%x" % (hsh)

    # if the endpoint is a filename, we use timestamp hashing
    # (but we should really use content hashing, i guess)
    if endpoint == 'static':
        filename = values.get('filename', None)
        if filename:
            file_path = os.path.join(app.root_path,
                                     endpoint, filename)
            values['q'] = int(os.stat(file_path).st_mtime)

    return flask.url_for(endpoint, **values)

def inject_components(response):
    if not flask.request.pudgy.pipelined:
        if response.headers["Content-Type"].find("text/html") == 0:
            if flask.request.pudgy.components:
                injection = marshal_components()
                response.set_data("%s\n%s" % (response.get_data(as_text=True), injection))

    return response

@simple_component.before_app_first_request
def install_pudgy():
    components.validate_components()


    install(flask.current_app)

    # because our first before_request handler never runs, we invoke it
    # manually
    add_components()

def install(app):
    global APP
    APP = app


    app.jinja_env.trim_blocks = True
    app.jinja_env.lstrip_blocks = True

    app.before_request(add_components)
    app.after_request(inject_components)
    app.jinja_env.globals.update(CC=render_component)

def register_blueprint(app):
    app.register_blueprint(simple_component)
